#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 12 15:20:43 2020
@author: jean
"""
import sys
import os
system_env = sys.argv[1] if len(sys.argv) > 1 else "jean"   # "eric
os.environ["system_env"] = system_env

import pickle 
#import Project_esnmpc.modelica_model as m
from cvxopt import solvers,matrix
import scipy.io as io
solvers.options['show_progress'] = False
solvers.options['reltol'] = 1e-7
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
#from torch.utils.tensorboard import SummaryWriter

from matplotlib import pyplot as plt
from matplotlib.pyplot import plot, subplot, show


class SimulationModel:
    """
    Template class for an arbitrary model used to simulate the real plant to test controller.
    Should be more sophisticated than the blackboxmodel type.
    """
    
    def __init__(self):
        pass
    
    def model_output(self,u):
        """
        outputs the model given an input, should return the model output. Note that
        the states should be an internal variable.
        :param u: input to the system. Dimension: 1d array of size PNMPCController.n_mv
        :type u: numpy.ndarray
        :returns: one step prediction given u and internal state. Dimension: 1darray of size PNMPCController.n_out
        :rtype: numpy.ndarray
        """
        pass

class BlackBoxModel:
    """
    Template class of the black box model. Interacts with the PNMPC Predictor through association.
    """

    def __init__(self):
        pass

    def get_current_state(self):
        """
        template function that should return the current model state
        """
        pass

    def set_current_state(self, a):
        """
        Template function that should be able to arbitrariely define the state
        :param a: The new values for the state vector.
        :type a: numpy.ndarray
        """
        pass

    def get_n_out(self):
        """
        Template function that should return the number of output variables in the identification model.
        """
        return 1

    def get_n_in(self):
        """
        Template function that should return the number of input variables in the identification model.
        """
        return 1

    def update(self, u):
        """
        Template function of the black-box model one-step update. Necessary for the PNMPC free response computation.
        Function should return output at one time step
        :param u: The control action input to the system. Dimension: (n_mv x 1) or 1d
        :type u: numpy.ndarray
        """
        pass

    def get_dydu(self, u, y):
        """
        Template function that should return the gradient between the output and the control action.
        """
        pass



def reg_least_squares(X,Y,reg): 
    """
    Analytical response to the least squares. Returns a matrix of collumn vectors. Internal function to the ESN input.
    :param X: Features/inputs of the model. dimension: (number of training data x number of features)
    :type X: numpy.ndarray
    :param Y: Real outputs of the system to fit the model. Dimension: (number of training data x number of outputs).
    :type Y: numpy.ndarray
    :param reg: Regularization parameter. Scalar.
    :type reg: float.
    :returns: The training weights. Dimension: (number of features x number of outputs)
    :rtype: numpy.ndarray
    """
        
        
    P = np.dot(X.T,X)
    R = np.dot(X.T,Y)
    theta = np.linalg.solve(P+reg*np.eye(P.shape[0],P.shape[1]),R)               
        
        
    return theta
    
def block_triang_create(n,N):
    """
    Creates a block triangular matrix using the identity matrix instead of the number 1. Used for the find_du function
    in the controller
    :param n: size of each block (size of the identity matrix).
    :type n: int
    :param N: Size of the block matrix (NxN blocks).
    :type N: int
    :returns: The resulting block-trangular matrix. (Dimension: N*n x N*n)
    :rtype: numpy.ndarray
    """
    
    ones_matrix = np.eye(n)
    block_eye = np.zeros([N*n,N*n])
    for i in range(N):
        for j in range(i,N):
            block_eye[n*j:n*j+n,n*i:n*i+n] = ones_matrix
    return block_eye

def feature_scaling(x,xmax,xmin):
    """
    Scales the matrix turning from [xmin,xmax] to [0,1].
    The function works through broadcasting xmax and mmin into x,
    so it does not matter if x is a single data or a concatenation of data
    :param x: vector/matrix to be scaled. Dimension: Either an 1d array (size of xmax/xmin)
    or (number of data examples x size of xmax or xmin).
    :type x: numpy.ndarray
    :param xmax: High-end of the scaling. Should be a 1 dimensional array of the same size as xmin.
    :type xmax: numpy.ndarray
    :param xmin: Low-end of the scaling. Should be a 1 dimensional array of the same size as xmax.
    :type xmin: numpy.ndarray
    :returns: The scaled form of x.
    :rtype: numpy.ndarray
    """
    
    return (x - xmin)/(xmax - xmin)

def feature_descaling(x,xmax,xmin,normalmode=True):
    """
    Scales back the matrix from [0,1] to [xmin,xmax].
    in the controller
    The function works through broadcasting xmax and mmin into x,
    so it does not matter if x is a single data or a concatenation of data
    :param x: vector/matrix to be scaled. Dimension: Either an 1d array (size of xmax/xmin)
    or (number of data examples x size of xmax or xmin).
    :type x: numpy.ndarray
    :param xmax: High-end of the scaling. Should be a 1 dimensional array of the same size as xmin.
    :type xmax: numpy.ndarray
    :param xmin: Low-end of the scaling. Should be a 1 dimensional array of the same size as xmax.
    :type xmin: numpy.ndarray
    :returns: The descaled form of x.
    :rtype: numpy.ndarray
    """

    return x*(xmax-xmin) + xmin

def change_or_not(x,min_val,max_val):
    y = 0
    p_change = np.random.rand()
    if p_change < 0.5:
        y = x
    else:
        y = min_val + (max_val - min_val)*np.random.rand()
    return y

def RFRAS(min,max,num_steps,minimum_step):
    """
    Actually the APRBS computation.
    :param min: low-end of the values the curve can assume.
    :type min: float
    :param max: high end of the values the curve can assume.
    :type max: float
    :param num_steps: size in time-steps of the signal.
    :type num_steps: float
    :param minimum_step: Minimum number of steps the signal will hode one values.
    :returns: The whole aprbs signal (1d array of size num_steps)
    :rtype: numpy.ndarray
    """
    RFRAS_sig = np.empty(num_steps)
    val = min + (max - min)*np.random.rand()
    for i in range(num_steps):

        if i % minimum_step  == 0:
            val = change_or_not(val,min,max)


        RFRAS_sig[i] = val

    return RFRAS_sig

def broadcast_vector(u,N):
    """
    Concatenates the 1d ndarray or collumn 2d ndarray u, N times.
    """
    broadcast = np.ones([u.shape[0]*N,1])
    for i in range(N):
        broadcast[u.shape[0]*i:u.shape[0]*i+u.shape[0]] = np.atleast_2d(u).reshape([u.size,1])
    return broadcast

def broadcast_diagonal_matrix(m,N):
    """
    Concatenates the 2d array m diagonally, N times.
    """
    
    broadcast = np.zeros([m.shape[0]*N,m.shape[1]*N])
    
    for i in range(N):
        broadcast[m.shape[0]*i:m.shape[0]*i+m.shape[0],m.shape[0]*i:m.shape[0]*i+m.shape[0]] = m
        
    return broadcast

def get_piece_of_vector(V,line_num,line_tot,start_line = 0):
    """
    Internal functon related to find_du. Obtain a new vector from a fraction of the input vector V.
    :param V: input vector, either one-dimensional or a collumn vector, size must be a multiple of line_tot.
    :type V: numpy.ndarray
    :param line_num: Num of elements selected from each line_tot sized block  of V. Smaller than Line_tot.
    :type line_num: float
    :param line_tot: Total size of each block in V.
    :type line_tot: float
    :param start_line: Start of the selection at each block. (start_line + line_num < line_tot.)
    :returns: Block vector, with blocks of size line_num. Has the same number of blocks as V.
    """
    
    V_new = np.empty([V.shape[0]*line_num//line_tot,1])
    
    for i in range(V_new.shape[0]//line_num):
        V_new[line_num*i:line_num*i+line_num] = \
            V[start_line+line_tot*i:start_line+line_tot*i+line_num]
            
    return V_new
        
def get_piece_of_matrix(G,line_num,col_num,line_tot,col_tot,start_line = 0,start_col = 0):
    """
    Analogue to the notion of get_piece of vector, however
    the returned blockmatrix contain blocks of size (line_num x col_num)
    and the input matrix G contains blocks of size (line_tox x col_tot).
    THe pruning beggins at line number start_lint and collumn number start_col inside each block.
    """
    
    G_new = np.empty([G.shape[0]*line_num//line_tot,G.shape[1]*col_num//col_tot])
#    if start_line > line_num - line_num/line_tot or start_col > col_num - col_num/col_tot:
#        print("Nao faz isso...")
    for i in range(G_new.shape[0]//line_num):
        for j in range(G_new.shape[1]//col_num):
            G_new[line_num*i:line_num*i+line_num,col_num*j:col_num*j + col_num] = \
            G[start_line+line_tot*i:start_line+line_tot*i+line_num,start_col + col_tot*j:start_col + col_tot*j + col_num]
        
    return G_new
  



class LSTM(nn.Module, BlackBoxModel):
    """
    LSTM class, should not be relevant as of now.
    """

    def __init__(self, input_dim, hidden_dim, output_dim=1,
                 num_layers=2,dropout = 0.0):
        super(LSTM, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers

        # Define the LSTM layer: num_layers >=2 defines a stacked LSTM
        self.lstm = nn.LSTM(self.input_dim, self.hidden_dim, self.num_layers,dropout = dropout).double()

        # Define the output layer
        self.linear = nn.Linear(self.hidden_dim, output_dim).double()

        self.init_hidden()

    def init_hidden(self):
        # This is what we'll initialise our hidden state and cell state as
        self.hidden_state = (torch.zeros(self.num_layers, 1, self.hidden_dim).double(),
                torch.zeros(self.num_layers, 1, self.hidden_dim).double())
        self.hidden_state[0].requires_grad = True
        self.hidden_state[1].requires_grad = True
        
        self.u = torch.zeros(self.input_dim,1)

    def forward(self, input):
        # Forward pass through LSTM layer
        # shape of lstm_out: [sequence length, batch_size, hidden_dim]
        # shape of self.hidden: (a, b), where a and b both
        # have shape (num_layers, batch_size, hidden_dim).

        #print("inp size ", input.size())
        #print(input.view(input.size(0), 1, -1).size())
        self.hidden_state_prev = self.hidden_state

        self.lstm_out, self.hidden_state = self.lstm(input.view(input.size(0), 1, -1), \
                                                     self.hidden_state_prev)

        # Only take the output from the final timetep
        # Can pass on the entirety of lstm_out to the next layer if it is a seq2seq prediction
        # y_pred = self.linear(lstm_out[-1].view(self.batch_size, -1))

        y_pred = self.linear(self.lstm_out.view(input.size(0), -1))

        #print("y_pred size ", y_pred.size())
        return y_pred.view(-1)

    def get_current_state(self):
        return self.hidden_state

    def set_current_state(self, a):
        self.hidden_state = a
        
    def get_current_output(self):
        y_pred = self.linear(self.hidden_state[0].view(1,-1))
        return y_pred.view(-1,1).detach().numpy()

    def get_n_out(self):
        return self.linear.out_features

    def get_n_in(self):
        return self.input_dim

    def update(self, u, isnumpy = True):
        # i = 0  # one instance/timestep
        u = torch.tensor(u)
        u.requires_grad = True
        u = u.view(1, 1, -1)
        y = self.forward(u)
        if isnumpy:
            return y.view(-1,1).data.numpy()  # .detach().numpy()
        else:
            return y.view(-1,1), u
    
    def get_dydu(self, u, y):
        n_out = len(y)
        jacT = torch.zeros(n_out,self.input_dim)
        for i in range(n_out):
            output = torch.zeros(n_out,1)
            output[i] = 1.
            jacT[i,:] = torch.autograd.grad(y, u, grad_outputs=output, retain_graph=True)[0]  # jacT[:, i:i + 1]
            #print(jacT[:, i:i + 1])
        return jacT.numpy()

    def get_dydu2(self, u, y):
        n_out = len(y)
        jacT = torch.zeros(2, n_out)
        for i in range(n_out):
            output = torch.zeros(n_out,1)
            output[i] = 1.
            jacT[:, i] = torch.autograd.grad(y, u, grad_outputs=output, retain_graph=True)[0]  # jacT[:, i:i + 1]
            #print(jacT[:, i:i + 1])
        return jacT.numpy()

    def train(self, x_data,y_data,valsize=0.25, num_epochs = 20):

        # n_train = 40000
       #n_samples = 1
        #scalings = dataset.FourTankSettings().scalings()

        #inputs, outputs, _ = dataset.four_tank(n_samples=n_samples, total_instances=n_train, scaling=scalings)
        #val_inputs, val_outputs, _ = dataset.four_tank(n_samples=1, total_instances=10000, start_from=n_train, scaling=scalings)
        data_size = x_data.shape[0]
        training_data_size = int(data_size*(1-valsize))
        val_data_size = int(data_size*valsize)
        inputs = x_data[:training_data_size]
        outputs = y_data[:training_data_size]
        
        val_inputs = x_data[training_data_size]
        val_outputs = y_data[training_data_size]
        

        seq_len = 200
        n_batches = int(inputs.shape[0]/seq_len)

        #model = LSTM(input_dim=input_dim, hidden_dim=20, output_dim=outputs.shape[1], num_layers=1)
        #################################
        loss_fn = torch.nn.MSELoss()  # size_average=False)
        learning_rate = 0.001
        optimiser = torch.optim.Adam(self.parameters(), lr=learning_rate)

        #####################
        # Train model
        #####################

        X_train = torch.tensor(inputs, dtype=torch.double).view(n_batches, seq_len, -1).transpose(0,1)
        y_train = torch.tensor(outputs, dtype=torch.double).view(n_batches, seq_len, -1).transpose(0,1)

        X_val = torch.tensor(val_inputs, dtype=torch.double).view(val_inputs.shape[0], 1, -1)
        y_val = torch.tensor(val_outputs, dtype=torch.double).view(val_inputs.shape[0], 1, -1)

        error = np.zeros(num_epochs)
        error_val = np.zeros(num_epochs)

        for t in range(num_epochs):
            loss_accum = 0.0
            loss_accum_val = 0.
            for k in range(n_batches):

                # Clear stored gradient
                self.zero_grad()

                # Initialise hidden state
                # Don't do this if you want your LSTM to be stateful
                self.init_hidden()  # TODO: check this: looks like it doesn't affect hidden state

                # Forward pass
                y_pred = self.forward(X_train[:,k,:])

                target = y_train[:,k]
                loss = loss_fn(y_pred, target.reshape(-1))
                loss_accum += loss.item()

                # Zero out gradient, else they will accumulate between epochs
                optimiser.zero_grad()
                loss.backward(retain_graph=False)
                optimiser.step()

            y_pred_val = self.forward(X_val).detach()
            loss_val = loss_fn(y_pred_val, y_val.reshape(-1))
            loss_accum_val += loss_val.item()

            error[t] = loss_accum
            error_val[t] = loss_accum_val
            print("Epoch ", t, "MSE: ", loss_accum, loss_accum_val,"gradient: ")
            gradnorm = 0.0
            count = 0.0
            for param in self.parameters():
                gradnorm += torch.norm(param.grad)**2
                count += 1.0
            gradnorm *= 1.0/count
            print(gradnorm)
            if gradnorm < 1e-8:
                break
            
        y_pred_val = self.forward(X_val).detach()
        loss_val = loss_fn(y_pred_val, y_val.reshape(-1))

        N = 1500
        i = 1
        #plot(y_val.detach().numpy()[:N,0,i],'-k')
        #plot(y_pred_val.view(-1,4).detach().numpy()[:N,i],'-.r')
        
        #plot(target.detach().numpy()[:N,i],'-k')
        #plot(y_pred.view(-1,4).detach().numpy()[:N,i],'-.r')
        #plt.show()
        
        
def sparsity(M, psi):
    """
    Explointing the point to point multiplication operator of python, this function
    merely turns a random number of elements in a matrix into zero.
    the probability of an element being zero is psi, and M is the input matrix.
    """
    N = np.empty_like(M)
    for linha in range(len(N)):
        for coluna in range(len(N[linha])):
            prob = np.random.rand()
            if prob < psi:
                N[linha][coluna] = 0
            else:
                N[linha][coluna] = 1


    return N*M

def grad_tanh(z):
    """
    Gradient of the elementwise atan.
    """
    
    
    return np.diag(1 - np.tanh(z.flatten())**2)

class EchoStateNetwork(BlackBoxModel):
    """
    Echo State Network class, child of the BlackBoxModel class.
    """

    def __init__(self,neu,n_in,n_out,
                 gama=0.5,ro=1,psi=0.5,in_scale=0.1,
                 bias_scale=0.5,alfa=10,forget = 1,
                 initial_filename="initial",
                 load_initial = False,save_initial = False,output_feedback = False, 
                 noise_amplitude = 0, 
                 out_scale = 0,default_warmupdrop=100):
        """
        The class initializes with the following parameters, all are scalars unless stated otherwise:
        :param neu: number of neurons.
        :param ro: spectral radius of the networks.
        :param psi: sparsity factor.
        :param in_scale: relative gan of the input to the reservoir and the reservoir to reservoir weights.
        :param bias_cale: relative gain of the bias.
        :param alfa: Initial Condition of the RLS algorithm.
        :param forget: forgetting factor of the RLS algorithm.
        :param initial_filename: file name where de weights of the network are to be saved.
        :type initial_filename: string
        :param load_initial: Will the weights be loaded from an already saved file?
        :type load_initial: bool
        :param save_initial: Do you wish to save the initial weights obtained randomly?
        :type save_initial: bool
        :output_feedback: will the network have feedback?
        :type output_feedback: bool
        :param noise_amplitude: if you desire the presence of state noise, this is the noise amplitude.
        :param out_scale: Scaling of the output feedback weights, in case output_feedback is enabled.
        :param default_warmupdrop: how much examples are to dropped each time data is acquired from a simulation.
        """
        #All matrixes are initialized under the normal distribution.
        print("initializing reservoir")
        self.neu = neu
        self.n_in = n_in
        self.n_out = n_out
        
        self.psi = psi #the network's sparcity, in 0 to 1 notation
        # Reservoir Weight matrix.
        self.Wrr0 = np.random.normal(0,1,[neu,neu])
        self.Wrr0 = sparsity(self.Wrr0, self.psi)
        # input-reservoir weight matrix
        self.Wir0 = np.random.normal(0,1,[neu,n_in])
        # bias-reservoir weight matrix
        self.Wbr0 = np.random.normal(0,1,[neu,1])

        self.Wor0 = np.random.normal(0,1,[neu,n_out])

        #self.Wbo = np.random.normal(0,1,[n_out,1])
        # reservoir-output weight matrix
        
        self.Wro = np.random.normal(0,1,[n_out,neu+1])
        #self.Wro = np.zeros([n_out,neu])

        self.leakrate = gama #the network's leak rate
        self.ro = ro #the network's desired spectral radius
        self.in_scale = in_scale #the scaling of Wir.
        self.bias_scale = bias_scale #the scaling of Wbr

        # learning rate of the Recursive Least Squares Algorithm
        self.alfa = alfa
        # forget factor of the RLS Algorithm
        self.forget = forget
        self.output_feedback = output_feedback

        #self.a = np.random.normal(0, 1, [neu, 1])
        self.a = np.zeros([neu, 1],dtype=np.float64)
        #save if save is enabled
        if save_initial:
            self.save_initial_fun(initial_filename)

        #load if load is enabled
        if load_initial:
            self.load_initial(initial_filename)

        # the probability of a memeber of the Matrix Wrr being zero is psi.
        self.Wrr = self.Wrr0
        #forcing Wrr to have ro as the maximum eigenvalue
        eigs = np.linalg.eigvals(self.Wrr)
        radius = np.abs(np.max(eigs))
        #normalize matrix
        self.Wrr = self.Wrr/radius
        #set its spectral radius to rho
        self.Wrr *= ro


        #scale tbe matrices
        self.Wbr = bias_scale*self.Wbr0
        self.Wir = in_scale*self.Wir0
        self.Wor = out_scale*self.Wor0
        
        
        #define the noise amplitude.
        self.noise = noise_amplitude


        #covariance matrix
        self.P = np.eye(neu+1)/alfa
        
        
        #variables related to the accumulation of data
        self.cum_data_input = np.array([]) #accumulated input data
        self.cum_data_output = np.array([]) #accumulated output data
        self.number_of_simulations = 0 # number of simulations
        self.simulations_start = [] #list that tracks each simulation start.
        self.default_warmupdrop = default_warmupdrop #default warumupdrop
        
        
    def get_n_out(self):
        """
        gets the number of outputs
        :returns: number of outputs in the network.
        :rtype: int
        """
        return self.n_out
    
    def get_n_in(self):
        """
        gets the number of inputs
        :returns: number of inputs in the network.
        :rtype: int
        """
        return self.n_in
    
    def get_current_state(self):
        """
        returns the current state of the network.
        :returns: network states, dimension (self.neu x 1)
        :rtype: numpy.ndarray
        """
        return self.a
    
    def set_current_state(self,a):
        """
        arbitrarierly defines the current network state.
        :param a: new network state, dimension: (self.neu x 1)
        :type a: numpy.ndarray
        """
        self.a = a

    def get_wro(self,n=0): 
        """
        returns a row of Wro given a certain output.
        :param n: outnput number desired.
        :type n: int
        :returns: Wro,row n.
        :rtype: numpy.ndarray
        """

        return self.Wro[n]

    def training_error(self,ref):
        """
        For RLS. given a desired output ref, computes the training error.
        """

        Ref = np.array(ref,dtype = np.float64)
        if self.n_out > 1:
            Ref = Ref.reshape(len(ref),1)

        e = np.dot(self.Wro,np.vstack((np.atleast_2d(1.0),self.a))) - Ref
        return e

    def train(self,ref):
        """
        RLS, NOT main training function. Given a certain desired output ref (dim self.n_out x 1),
        updates the training weights self.Wro.
        """
        e = self.training_error(ref)
        #the P equation step by step
        a_wbias = np.vstack((np.atleast_2d(1.0),self.a))
        self.P = \
        self.P/self.forget - np.dot(np.dot(np.dot(self.P,a_wbias),a_wbias.T),self.P)/ \
        (self.forget*(self.forget + np.dot(np.dot(a_wbias.T,self.P),a_wbias)))
        #self.sigma_q = (1 - 1 / (self.K_a * self.neu)) * \
        #               self.sigma_q + \
        #               (1 - (1 - 1 / (self.K_a * self.neu))) * \
        #               (np.dot(np.dot(self.a.T, self.P), self.a)) \
        #               ** 2
        for saida in range(self.n_out):

            #Transpose respective output view..
            Theta = self.Wro[saida,:]
            Theta = Theta.reshape([self.neu+1,1])


            #error calculation
            Theta = Theta - e[saida]*np.dot(self.P,a_wbias)
            Theta = Theta.reshape([1,self.neu+1])

            self.Wro[saida,:] = Theta




    def update(self,inp,y_in = np.atleast_2d(0),training = False ):
        """
        State updating function of the network. Ignore y_in if you are not using output_feedback.
        :param inp: input to the network (self.n_in x 1) or 1d array (self.n_in)
        :type inp: numpy.ndarray
        :param y_in: output feedback, dim (self.n_out x 1).
        :type y_in: numpy.ndarray
        :param training: The function would like to know if you are using update to gather data for training (internal functions only).
        :type training: bool
        :returns: network output. 
        """
        # input has to have same size
        # as n_in. Returns the output as shape (2,1), so if yo
        # u want to plot the data, a buffer is mandatory.
        Input = np.array(inp)
#        print("u0",Input)
        Input = Input.reshape(Input.size,1)
#        print("u1",Input)
        Y_in = np.array(y_in)
#        print("y0",Y_in)
        Y_in = Y_in.reshape(Y_in.size, 1)
#        print("y1",Y_in)
#        print("Inpitsize",Input.size)
#        print("n_in",self.n_in)
        if (y_in == 0).all():
            Y_in = np.zeros([self.n_out,1])
        if Input.size == self.n_in:                
            z = np.dot(self.Wrr,self.a) + np.dot(self.Wir,Input) + self.Wbr
            if self.output_feedback:
                z += np.dot(self.Wor,Y_in)
            if self.noise > 0 and training:
                z += np.random.normal(0, self.noise, [self.neu, 1])                
            self.a = (1-self.leakrate)*self.a + self.leakrate*np.tanh(z)
            
            a_wbias = np.vstack((np.atleast_2d(1.0),self.a))
            y = np.dot(self.Wro,a_wbias)
            return y
        else:
            raise ValueError("input must have size n_in")

    def copy_weights(self, outer_esn):
        """
        Receives weights from another ESN.
        """
        if self.Wro.shape == outer_esn.Wro.shape:
            self.Wro = np.copy(outer_esn.Wro)
            self.Wrr = np.copy(outer_esn.Wrr)
            self.Wir = np.copy(outer_esn.Wir)
            self.Wbr = np.copy(outer_esn.Wbr)
        else:
            print("shapes of the weights are not equal")

    def save_reservoir(self,fileName):
        """
        Save current reservoir into file named string fileName.
        """
        data = {}
        data['Wrr'] = self.Wrr
        data['Wir'] = self.Wir
        data['Wbr'] = self.Wbr
        data['Wro'] = self.Wro
        data['a0'] = self.a
        io.savemat(fileName,data)

    def load_reservoir(self,fileName):
        """
        Load reservoir from fileName.
        """
        data = {}
        io.loadmat(fileName, data)
        self.load_reservoir_from_array(data)

    def load_reservoir_from_array(self, data):
        """
        Load reservoir given a dictionary with the following keys:
        Wrr for reservoir to reservoir weights.
        Wir for input to reservoir weights.
        Wbr for bias to reservoir weights.
        Wro for reservoir to output weights.
        """
        self.Wrr = data['Wrr']
        self.Wir = data['Wir']
        self.Wbr = data['Wbr']
        self.Wro = data['Wro']
        if 'a0' in data:  # check by Eric
            self.a = data['a0']
            
        # added by Eric - start
        if 'Wro_b' in data:
            self.Wro = np.hstack((data['Wro_b'], self.Wro))

        if 'leak_rate' in data:
            try:
                self.leakrate = data['leak_rate'][0][0]
            except:
                self.leakrate = data['leak_rate']
        self.neu = self.Wrr.shape[0]
        self.n_in = self.Wir.shape[1]
        self.n_out = self.Wro.shape[0]            
        # added by Eric - end
        

    def load_initial(self,filename):
        """
        Load initial conditions from a filename.
        """
        data = {}
        print("loading reservoir")
        io.loadmat(filename, data)
        self.Wrr0 = data['Wrr']
        self.Wir0 = data['Wir']
        self.Wbr0 = data['Wbr']
        self.Wro = data['Wro']
        #self.Wor0 = data['Wor']
        self.a = data['a0']

    def save_initial_fun(self,filename):
        """
        Save initial conditions to a filename.
        """
        data = {}
        data['Wrr'] = self.Wrr0
        data['Wir'] = self.Wir0
        data['Wbr'] = self.Wbr0
        data['Wor'] = self.Wor0
        data['Wro'] = self.Wro
        data['a0'] = self.a
        print( "saving reservoir")
        io.savemat(filename, data)

    
#    def trainLMS(self,ref):
#
#        learningrate = 1
#
#        e = self.trainingError(ref)
#        for saida in range(self.n_out):
#            Theta = self.Wro[saida, :]
#            Theta = Theta.reshape([self.neu, 1])
#            Theta = Theta - learningrate*e*self.a/np.dot(self.a.T,self.a)
#            self.Wro[saida, :] = Theta.T
#

    def offline_training(self,X,Y,regularization,warmupdrop): #X is a matrix in which X[i,:] is all parameters at time i. Y is a vector of desired outputs.
        """
        Simplest form of training, with an outsidely defined warmupdrop.
        Follows the same logic areg_least_squares. I do not recommend the use of this function. 
        """
        A = np.empty([Y.shape[0]-warmupdrop,self.neu])
        for i in range(Y.shape[0]):
            if self.output_feedback:
                if i > 0:
                    self.update(X[i,:],Y[i-1,:])
                else:
                    self.update(X[i,:])
                if i > warmupdrop:
                    A[i-warmupdrop,:] = self.a.T
            else:
                self.update(X[i, :])
                if i > warmupdrop:
                    A[i - warmupdrop, :] = self.a.T
                    
                    
        A_wbias = np.hstack((np.ones([A.shape[0],1]),A))

        P = np.dot(A_wbias.T,A_wbias)
        R = np.dot(A_wbias.T,Y[warmupdrop:])
            #print R,"R"
        Theta = np.linalg.solve(P+regularization*np.eye(self.neu+1,self.neu+1),R)
        self.Wro = Theta.T
        
        
        
        
    def add_data(self,input_data,output_data,warmupdrop):
        """
        Adds data to the class database. Run this function after a single simulation. The purpose of this function
        is to accumulate data from many different simulations using the same system.
        :param input_data: data with the input, (number of data x self.n_in)
        :type input_data: numpy.ndarray
        :param output_data: data with the desired output (number of data x self.n_out)
        :type output_data: numpy.ndarray
        :param warmupdrup: the number of data that shall be dropped when gathering data
        """
        #warning, always handle both cum_Data variables in pair. THis function adds data for the ESN to train.
        
        
        self.simulations_start.append(self.cum_data_input.shape[0])
        A = np.empty([input_data.shape[0]-warmupdrop,self.neu])
        for i in range(input_data.shape[0]):
            if self.output_feedback:
                if i > 0:
                    self.update(input_data[i,:],output_data[i-1,:],training = True)
                else:
                    self.update(input_data[i,:],training = True)
                if i > warmupdrop:
                    A[i-warmupdrop,:] = self.a.T
            else:
                self.update(input_data[i, :],training = True)
                if i > warmupdrop:
                    A[i - warmupdrop, :] = self.a.T
        if self.cum_data_input.size == 0:
            self.cum_data_input = A
            self.cum_data_output = output_data[warmupdrop:]
            
        else:
            self.cum_data_input = np.vstack([self.cum_data_input,A])
            self.cum_data_output = np.vstack([self.cum_data_output,output_data[warmupdrop:]])
            
        self.number_of_simulations += 1
        self.reset()
        

    
    
    def cum_train_cv(self,min_reg,max_reg,tests=50,folds = 10): #use after using add_data to add data.
        """
        main training function, use after at least one instance of add_data. Comes with built-in cross validation. Float unless
        stated otherwise
        :param min_reg: minimum regularization value.
        :param max_reg: maximum regularization value.
        :param reg_test: number of tests made. Default at 50.
        :type reg_test: int
        :param folds: number of folds in the cross-validation. Defaults at 10.
        :type folds: int
        :returns: the resulting training error of the cv, and the best regularization parameter.
        :rtype: tupĺe
        """
        #crossvalidation startup
        reg_list = np.linspace(min_reg,max_reg,tests) #regularization parameter candidates
        error_list = np.empty(folds) #list of errors.
        val_size = self.cum_data_input.shape[0]//folds #size of each subset.
        
        A_wbias = np.hstack((np.ones([self.cum_data_input.shape[0],1]),self.cum_data_input)) #add bias

        
        best_error = 9999999999999999999.0 ## comparator
        best_reg = 0.0
        
        
        
        for i,regularization in enumerate(reg_list): 
            
            for fold in range(folds):
                #separate training set and validation set
                if fold < folds - 1: #separate folds
                    training_A = np.vstack([A_wbias[:val_size*fold,:],A_wbias[val_size*fold+val_size:,:]])                
                    training_y = np.vstack([self.cum_data_output[:val_size*fold,:],
                                            self.cum_data_output[val_size*fold+val_size:,:]])
                else:
                    training_A = A_wbias[:val_size*fold,:]
                    training_y = self.cum_data_output[:val_size*fold,:]
                
                cv_A = A_wbias[val_size*fold:val_size*fold+val_size,:]
                cv_y = self.cum_data_output[val_size*fold:val_size*fold+val_size,:] 
                
                #run training
                theta = reg_least_squares(training_A,training_y,regularization)
                error_list[fold] =  np.mean((cv_y - np.dot(cv_A,theta))**2)
                
            mean_error = np.mean(error_list)
            if mean_error < best_error:
                best_error = mean_error
                best_reg = reg_list[fold]
            
        
        
        
    
        self.Wro = reg_least_squares(A_wbias,self.cum_data_output,best_reg).T
        
        
        #the functon returns the cv error of the weights used.
        return best_error,best_reg     

    def reset(self):
        """
        Resets the networks bringing the state into 0.
        """

        self.a = np.zeros([self.neu, 1])

    def get_forgetingfactor(self):

        return self.forget
    
    def get_state_from(self,outer_esn):
        """
        Copy state from another network
        """
        
        self.a = outer_esn.a
        
    def get_derivative_df_du(self, inp,y_in = np.atleast_2d(0)):
        """
        Obtains dfdu gradient, for gradient calculation purposes.
        """
        
        Input = np.array(inp)
        Input = Input.reshape(Input.size,1)
        #Y_in = np.zeros([self.n_out,1])
        z = np.dot(self.Wrr,self.a) + np.dot(self.Wir,Input) + self.Wbr
        if not (y_in == 0).all():
            Y_in = np.array(y_in)
            Y_in = Y_in.reshape(Y_in.size, 1)
            z += np.dot(self.Wor,Y_in)
                     
        J = grad_tanh(z)
        
        return np.dot(J,self.Wir)
    
    def get_derivative_df_dx(self, inp,y_in = np.atleast_2d(0)):
        """
        obtains dfdx gradient, for gradient calculation purposes.
        """
        
        Input = np.array(inp)
        Input = Input.reshape(Input.size,1)

        z = np.dot(self.Wrr,self.a) + np.dot(self.Wir,Input) + self.Wbr
        if not (y_in == 0).all():
            Y_in = np.array(y_in)
            Y_in = Y_in.reshape(Y_in.size, 1)
            z += np.dot(self.Wor,Y_in)
        
        J = grad_tanh(z)

        z1 = self.Wrr
        if self.output_feedback:
            z1 += np.dot(self.Wor,self.Wro[:,1:])
        z1 = np.dot(J,z1)
        
        return (1-self.leakrate)*np.eye(self.neu) + self.leakrate * z1
        
    def get_current_output(self):
        """
        Returns the output of the network.
        """
        a_wbias = np.vstack((np.atleast_2d(1.0),self.a))
        return np.dot(self.Wro,a_wbias)
    
    def covariance_reset(self,diag_alpha):
        self.P = np.eye(self.neu)/diag_alpha
    

    @staticmethod
    def new_rnn_from_weights(weights):
        esn_4tanks = EchoStateNetwork(
            neu = 400,
            n_in = 2,
            n_out = 3,
            gama = 0.1,
            ro = 0.99,
            psi = 0.0,
            in_scale = 0.1,
            bias_scale = 0.1,
            initial_filename = "4tanks1",
            load_initial = False,
            save_initial = False,
            output_feedback = False)
        esn_4tanks.load_reservoir_from_array(weights)
        esn_4tanks.reset()
        return esn_4tanks
    

class PNMPCPredictor:
    
    def __init__(self,control_horizon,prediction_horizon,bb_model = BlackBoxModel(),filter_constants = 0.3, K = 1.0):
        """
        Class reprenting the PNMPC predictor.
        :param control_horizon: control horizon of the MPC.
        :type control_horizon: int
        :param prediction_horizon: prediction horizon of the MPC.
        :type prediction_horizon: int.
        :param bb_model: Black-Box model used by the PNMPC, this is where the ESN comes in.
        :type bb_model: BlackBoxModel
        :param filter_constants: time constant of the filter. Scalar.
        :param K: integrator gain.
        """
        #G matrix.
        self.dynamic_matrix = np.zeros([prediction_horizon*bb_model.get_n_out(),control_horizon*bb_model.get_n_in()])
        
        self.F = np.empty([prediction_horizon*bb_model.get_n_out(),1])
        
        
        self.bb_model = bb_model
        
        self.n_y = prediction_horizon
        self.n_u = control_horizon
        
        self.correction = np.zeros([bb_model.get_n_out(),1])
        self.integral_error = np.zeros([bb_model.get_n_out(),1])
        
        self.filter_constants = filter_constants*np.eye(bb_model.get_n_out())
        self.K = K

    def calculate_matrices(self, u, y_real):
        """
        Based on the input and real output, it calculates the matrices.
        :param u: Control Action, dimension (self.bb_model.get_n_in() x 1)
        :param y_real: real plant output, (self.bb_model.get_n_out() x 1)
        """
        if self.bb_model.__class__.__name__ == "EchoStateNetwork" or self.bb_model.__class__.__name__ == "RecurrentNeuralNetwork":
            self.calculate_matrices_split( u, y_real)
        elif self.bb_model.__class__.__name__ == "LSTM":
            self.calculate_matrices_dydu( u, y_real)

    def calculate_matrices_dydu(self, u, y_real):
        # self.erro = y_real - y_predito + np.dot(self.filter_constants,self.erro)
        # self.erro_integral = self.erro_integral + self.erro
        # dxdu = np.zeros([self.n_u,a.shape[0],self.rede.get_n_in()])
        
        #calculation of the pure F and G
        a = self.bb_model.get_current_state()
        d0 = self.bb_model.get_n_out()
        d1 = self.bb_model.get_n_in()
        input_sequence = dict()
        for i in range(0, self.n_y): # horizonte de predicao
            output, input = self.bb_model.update(u,isnumpy=False)
            input_sequence[i] = input  # input used in the prediction horizon indexed by 'i'            
            self.F[i * d0:i * d0 + d0] = output.data.numpy()   # TODO: make this general
            for j in range(0, i + 1):  # horizonte de controle
                if j < self.n_u:
                    # dxdu[j] = J + np.dot(S,dxdu[j])
                    self.dynamic_matrix[d0 * i:d0 * i + d0, d1 * j:d1 * j + d1] = self.bb_model.get_dydu(input_sequence[j], output)
                else:
                    pass
        self.bb_model.set_current_state(a)
        
        
        #calculation of the correction factor
        y_real = np.array(y_real).reshape([y_real.size, 1])       
        y_pred = self.bb_model.get_current_output() + self.integral_error
        error = y_real - y_pred
        self.correction = np.dot(np.eye(self.bb_model.get_n_out()) - self.filter_constants, error) + np.dot(
            self.filter_constants, self.correction)
        # print self.correcao,"valor do fator de correcao"
        self.integral_error = self.integral_error + self.K * self.correction
        # -----------------
        
        
        self.F = self.F + broadcast_vector(self.integral_error, self.n_y)

    def calculate_matrices_split(self,u,y_real):
        y_real = np.array(y_real).reshape([y_real.size,1])
        y_pred = self.bb_model.get_current_output() + self.integral_error
        #self.rede.train(y_real)
        #---state space approach
        error = y_real - y_pred 
        self.correction = np.dot(np.eye(self.bb_model.n_out)-self.filter_constants,error) + np.dot(self.filter_constants,self.correction)
        #print self.correcao,"valor do fator de correcao"
        self.integral_error = self.integral_error + self.K*self.correction
        #print self.erro_integral, "integral do erro"
        #print self.erro_integral
        #self.rede.a = np.dot(self.L,erro) + self.rede.a
        #-----------------
        a = self.bb_model.get_current_state()
        #self.erro = y_real - y_predito + np.dot(self.filter_constants,self.erro)
        #self.erro_integral = self.erro_integral + self.erro
        dxdu = np.zeros([self.n_u,self.bb_model.neu,self.bb_model.n_in])
        d0 = self.bb_model.n_out
        d1 = self.bb_model.n_in
        for i in range(0,self.n_y):
            self.F[i*d0:i*d0 + d0] = self.bb_model.update(u)
            J = self.bb_model.get_derivative_df_du(u)
            S = self.bb_model.get_derivative_df_dx(u)
            for j in range(0,i+1):
                if j < self.n_u:
                    dxdu[j] = J + np.dot(S,dxdu[j])
                    self.dynamic_matrix[d0*i:d0*i+d0,d1*j:d1*j+d1] = np.dot(self.bb_model.Wro[:,1:],dxdu[j])
                else:
                    pass
            
        self.bb_model.set_current_state(a)
        self.F = self.F + broadcast_vector(self.integral_error,self.n_y)
        
        
       
        
        return None
    
    def prediction(self,u,du,y_real): #Makes a Prediction given a certain deltaU.
        """
        Single Step prediction.
        """
        self.calculate_matrices(u, y_real)
        Y = np.dot(self.dynamic_matrix,du) + self.F
        
        
        return Y
    
    def update_network(self,u):
        """
        Update the network directly from the PNMPC class.
        """
        return self.bb_model.update(u)
        
    def get_current_prediction(self):
        """
        Returns current network output.
        """
        return self.bb_model.get_current_output()
        
    def test_prediction_proximity(self):
        """
        Debug function. prints the prediction error between the taylor approximation and the network.
        """
        
        du = 0.01*np.ones(self.n_u*self.bb_model.get_n_in())
        u0 = np.array([0.5,0.5])
        y_real = np.array([0.0,0.0,0.0,0.0])
        self.calculate_matrices(u0, y_real)

        y_pred_taylor = self.F + np.dot(self.dynamic_matrix,np.atleast_2d(du).T)
        
        y_pred = np.empty_like(y_pred_taylor)
        a = self.bb_model.get_current_state()

        n_in  = self.bb_model.get_n_in()
        n_out  = self.bb_model.get_n_out()

        u = u0
        for i in range(self.n_y):
            if i < self.n_u:
                u = u + du[n_in*i : n_in*(i + 1)]
            y_pred[n_out*i : n_out*(i+1)] = self.bb_model.update(u)

        self.bb_model.a = a
        print("predicted by taylor------------------", y_pred_taylor.transpose())
        print("predicted by black box model-------------------", y_pred.transpose())
        print("error-------------------", (y_pred_taylor - y_pred).transpose())

        for i in range(self.n_y):
            for j in range(self.n_u):
                print(self.dynamic_matrix[n_out * i:n_out * (i+1), n_in * j : n_in * (j+1)],
                      "jacobian matrix in instant k +", i, "with respect to u[k+", j, "]")
        
        
    def online_train_rls(self,ref):
        self.bb_model.train(ref)
        
        
        
    

class PNMPCController:
    
    def __init__(self,n_cv,
                 u_max,
                 u_min,
                 y_max,
                 y_min, 
                 initial_u,
                 Q,
                 R,
                 du_max_c,
                 output_constraints_max,
                 output_constraints_min,
                 predictor = PNMPCPredictor(3,20),
                 simulation_model = SimulationModel()):
        
        
        """
        Class that automates the process of a single simulation run, based on an already trained black box model, the pnmpc
        and a simulation model. ndarray unless stated otherwise.
        :param u_max: Upper-end of the plant input scaling. Should double as saturation constraint. 1d array of size self.bb_model.n_in
        :param u_min: Lower-end of the plant input scaling. Should double as saturation constraint. 1d array of size self.bb_model.n_in 
        :param y_max: Upper-end of the plant output scaling. 1d array of size self.n_out
        :param y_min: Lower-end of the plant output scaling. 1d array of size self.n_out
        :param initial_u: Initial condition of the run. 1darray of size self.bb_model.n_in
        :param Q: diagonal error penalizing weight matrix. size (n_cv x n_cv)
        :param R: diagonal input variation penalizing weight matrix. size (self.n_mv x self.n_mv)
        :param du_max_C: Constraint of maximum control variation. Can be either a scalar os a 1d array.
        :param output_constraints_max: maximum output constraints. 1d array of size (self.n_out - self.n_cv)
        :param output_constraints_min: minimum output constraints. 1d array of size (self.n_out - self.n_cv)
        :param predictor: PNMPC predictor.
        :type predictor: PNMPCPRedictor
        :param simulation_model: Model used to simulate the plant.
        :type simulation_model: SimulationModel
        """
        
        self.n_cv = n_cv #number of controlled variables. The first n_cv variables in the model output will be considered
        #controlled variables. 
        
        self.simulation_model = simulation_model
        self.pnmpc = predictor
        
        self.n_mv = self.pnmpc.bb_model.get_n_in()
        print("n_mn=",self.n_mv)
        self.n_out = self.pnmpc.bb_model.get_n_out()
        
        
        #the code considers application of feature scaling, so these variables are related to the scaling of the ESN training.
        self.u_max = u_max
        self.y_max = y_max
        self.u_min = u_min
        self.y_min = y_min
        
        
        #controller simulation run related
        
        self.initial_u = initial_u
        self.u_normal = feature_scaling(self.initial_u,self.u_max,self.u_min)
        
        
        
        
        #This part is related to the optimization problem:
        
        self.Q_ext = broadcast_diagonal_matrix(Q,self.pnmpc.n_y)
        self.R_ext = broadcast_diagonal_matrix(R,self.pnmpc.n_u)
        
        
        #define control variation
        
        if np.isscalar(du_max_c) == True:
            du_max = du_max_c*np.ones(self.n_mv)
            
        elif du_max_c.__class__.__name__ == 'ndarray':
            du_max = du_max_c
            
        else:
            raise Exception("du_max must be a numpy array or a scalar")
            
        du_min = -du_max
        
        self.b_dumax = broadcast_vector(du_max,self.pnmpc.n_u)
        self.b_dumin = broadcast_vector(du_min,self.pnmpc.n_u)
        
        
        #We consider that the model is already scaled under the saturation constraints
        u_normal_max = np.ones(self.n_mv)
        u_normal_min = np.zeros(self.n_mv)
        self.b_umax = broadcast_vector(u_normal_max,self.pnmpc.n_u)
        self.b_umin = broadcast_vector(u_normal_min,self.pnmpc.n_u)
        
        #Input-related constraints
        I = np.eye(self.pnmpc.bb_model.n_in*self.pnmpc.n_u)
        T = block_triang_create(self.pnmpc.bb_model.n_in,self.pnmpc.n_u)
        self.A_default = np.concatenate((I,-I,T,-T))
        
        y2_max = output_constraints_max
        y2_min = output_constraints_min
        
        #y1_max_normal = feature_scaling(y1_max,y_max[0:2],y_min[0:2])
        y2_max_normal = feature_scaling(y2_max,self.y_max[self.n_cv:],self.y_min[self.n_cv:])
        #y1_min_normal = feature_scaling(y1_min,y_max[0:2],y_min[0:2])
        y2_min_normal = feature_scaling(y2_min,self.y_max[self.n_cv:],self.y_min[self.n_cv:])
        
        
        #y1_max_nb = broadcast_vector(y1_max_normal,predictor.n_y)
        #y1_min_nb = broadcast_vector(y1_min_normal,predictor.n_y)
        self.y2_max_nb = broadcast_vector(y2_max_normal,self.pnmpc.n_y)
        self.y2_min_nb = broadcast_vector(y2_min_normal,self.pnmpc.n_y)
            
        
    def find_du(self,u,ref):
        
        """
        Function that computes the optimization problem, based on the current input, and the reference signal.
        :param u: Current control action  dim:(self.n_mv x1) or (self.n_mv)
        :param ref: Ref signal. dim: (self.n_cv) or (self.n_cv x 1)
        :returns: The control variation outputed by cvxopt.
        :rtype: numpy.ndarray
        """
        #weights of the quadratic function
    


        #Cost-function related, separate piece of prediction used for cost function, and piece of prediction used for constraints
        G1 =  get_piece_of_matrix(self.pnmpc.dynamic_matrix,self.n_cv,self.n_mv,self.n_out,self.n_mv)
        F1 = get_piece_of_vector(self.pnmpc.F,self.n_cv,self.n_out)
        
        Q = np.dot(G1.T,np.dot(self.Q_ext,G1)) +\
        self.R_ext
        ref_b = broadcast_vector(ref,self.pnmpc.n_y)
        p = np.dot(G1.T,np.dot(self.Q_ext,F1 - ref_b))
    
        #Input-related constraints
        u_const = broadcast_vector(u,self.pnmpc.n_u)
        h = np.concatenate([self.b_dumax,-self.b_dumin,self.b_umax - \
                            u_const,u_const - self.b_umin])
    
        #output related constraints
        try:
            G2 =  get_piece_of_matrix(self.pnmpc.dynamic_matrix,(self.n_out - self.n_cv),self.n_mv,self.n_out,self.n_mv,self.n_cv)
            F2 = get_piece_of_vector(self.pnmpc.F,(self.n_out - self.n_cv),self.n_out,self.n_cv)
            #A = np.concatenate((A,G1,-G1,G2,-G2))
            A = np.concatenate((self.A_default,G2,-G2))
        except:
            A = self.A_default
        
        finally:
            pass
    
#        h = np.concatenate((h,y1_max_nb - F1,-(y1_min_nb - F1),y2_max_nb - F2,-(y2_min_nb - F2)))
#        print("h=", h)
#        print("y2max = ", F2)
#        print("y2min", -(self.y2_min_nb))
        h = np.concatenate((h,self.y2_max_nb - F2,-(self.y2_min_nb - F2)))
#        print("TEST")
    
    
    
        #famoso "jogar no solver"
        Q_qp = matrix(2.0*Q)
        p_qp = matrix(p)
        h_qp = matrix(h)
        G_qp = matrix(A)
        sol=solvers.qp(Q_qp, p_qp, G_qp, h_qp)
        
        du = np.array(sol['x'])
        #print du
        #test purpose line
        #print self.test_prediction_proximity(du),"oi"
        #end test purpose
        
        #drop the non-imediate calculated dus.
        du = du[:self.pnmpc.bb_model.n_in]

        
        
        return du
        

    def run_controller(self,y_ref):
        """
        Based on a reference signal, run a simulation.
        Returns a dictionary with the following keys:
        y_plot: The system response.
        y_ref: the used reference signal.
        simtime: The simulation time.
        y_pred_plot: A priori prediction (correctionless)
        y_pred_posterior_plot After correction plot.
        correction_plot: A vector containing the correction factor along the runtime.
        IAE: The integral absolute error of the run.
        du_total: The total control variation of the run.
        """
        warm_up_time = 200
        simtime = y_ref.shape[0]
        y_ref_normal = feature_scaling(y_ref,self.y_max[:self.n_cv],self.y_min[:self.n_cv])
        u_plot = np.empty([simtime,self.n_mv])
        y_plot = np.empty([simtime,self.n_out])
        y_pred_plot = np.empty_like(y_plot)
        y_pred_posterior_plot = np.empty_like(y_plot)
        u_normal = self.u_normal
        for i in range(warm_up_time):
            y = self.simulation_model.model_output(self.initial_u)
            y_pred_normal = self.pnmpc.update_network(u_normal)

        
        correction_plot = np.empty([simtime,self.n_out])
        for i in range(simtime):
            u = feature_descaling(u_normal,self.u_max,self.u_min)
            u_plot[i] = u.flatten()
            y = self.simulation_model.model_output(u)
#            print(y)
            y_normal = feature_scaling(y,self.y_max,self.y_min)
            self.pnmpc.calculate_matrices(u_normal,y_normal)
            du = self.find_du(u_normal,y_ref_normal[i])
            u_normal = u_normal + du.flatten()
            y_pred_normal = self.pnmpc.update_network(u_normal)
            y_pred_normal_post = y_pred_normal + self.pnmpc.integral_error
            y_pred = feature_descaling(y_pred_normal.flatten(),self.y_max,self.y_min)
            y_pred_posterior = feature_descaling(y_pred_normal_post.flatten(),self.y_max,self.y_min)
            y_plot[i] = y.flatten()
            y_pred_plot[i] = y_pred
            y_pred_posterior_plot[i] = y_pred_posterior
            #print i,u_plot[i],"current input",y,"current output",Model.measurements[i]
            correction_plot[i] = self.pnmpc.integral_error.flatten()
            if i%100 == 0:
                print(i)
            if i%100 == 0:
                print(i,u_plot[i],"current input",y,"current output",y_ref[i],"current ref")
        



        #metricas finais de controle. Integral Absolute Error e Total Control Variation.
        IAE = np.sum(np.abs(y_ref - y_plot[:,:self.n_cv]),axis = 0)
        du_total = np.sum(np.abs(u_plot[1:] - u_plot[:-1]),axis = 0)
        save_dict = {'u_plot':u_plot,
             'y_plot':y_plot,
             'y_ref':y_ref,
             'simtime':simtime,
             'y_pred_plot':y_pred_plot,
             'correction_plot':correction_plot,
             'y_pred_posterior_plot':y_pred_posterior_plot,
             'IAE':IAE,
             'du_total':du_total}

#        savefile = open("control_plot.mat","w")
#        io.savemat(savefile,save_dict)
#
#        print("")
#        print("IAE:", IAE, y_ref.shape)
#        print("du_total:", du_total, u_plot.shape)
        
        return save_dict